﻿Imports System.Net.Sockets
Imports System.Text
Imports System.Text.RegularExpressions


''' <summary>
''' The Remote_K8000 class contains functionality to control the K8000 card over a network connection. 
''' A connection should be made with a TCP/IP server that supports this protocol.
''' 
''' Initially I wrote this class to be used with the GHI Electronics FEZ Panda II module, but it can
''' also work with other server software.
''' </summary>
''' <remarks></remarks>
Public Class Remote_K8000
    ' private class variables. Should be accessed only through the class properties
    Private _isConnected As Boolean
    Private _lastException As Exception

    ' connection objects
    Private tcpClient As New System.Net.Sockets.TcpClient()
    Private networkStream As NetworkStream

    ' This event will be called when a command is transmitted when no connection is available
    Public Event noConnectionError()

    ' This event will be called when the class detects the client is disconnected
    Public Event clientDisconnected()

    ' The controller is a timer that will check the connection
    Private WithEvents controller As Timer
    Private isSending As Boolean = False




#Region "Class Properties"
    ''' <summary>
    ''' the isConnected property holds if the system is currently connected or not.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property isConnected As Boolean
        Get
            Return _isConnected
        End Get
        Set(ByVal value As Boolean)
            _isConnected = value
        End Set
    End Property



    ''' <summary>
    ''' The lastException property contains the exception that last occured. It is used to get 
    ''' information about the problem.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property lastException As Exception
        Get
            Return _lastException
        End Get
    End Property
#End Region





#Region "K8000 Interface Commands"
    ''' <summary>
    ''' setAllIO() will transmit the command to set all the I/O channels. The server
    ''' response is not processed.
    ''' </summary>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function setAllIO() As Boolean
        Dim response As String = ""
        Return sendCommand("SETALLIO", response)
    End Function



    ''' <summary>
    ''' setAllIO() will transmit the command to set all the I/O channels. The server
    ''' response is processed into the IOdata array (boolean with the state of every
    ''' I/O channel.
    ''' </summary>
    ''' <param name="IOdata">The IOdata array contains the server response</param>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function setAllIO(ByRef IOdata() As Boolean) As Boolean
        Dim IOstring As String = ""
        Dim serverResponse As String = ""

        serverResponse = sendCommand("SETALLIO", IOstring)

        If serverResponse Then
            getIOconfigFromServerResponse(IOstring, IOdata)
            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' clearAllIO() will transmit the command to clear all the I/O channels. The server
    ''' response is not processed.
    ''' </summary>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function clearAllIO() As Boolean
        Dim response As String = ""
        Return sendCommand("CLEARALLIO", response)
    End Function



    ''' <summary>
    ''' clearAllIO() will transmit the command to clear all the I/O channels. The server
    ''' response is processed.
    ''' </summary>
    ''' <param name="IOdata"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function clearAllIO(ByRef IOdata() As Boolean) As Boolean
        Dim IOstring As String = ""
        Dim serverResponse As String = ""

        serverResponse = sendCommand("CLEARALLIO", IOstring)

        If serverResponse Then
            getIOconfigFromServerResponse(IOstring, IOdata)
            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' setDA() will set a D/A channel
    ''' </summary>
    ''' <param name="channel">the channel to change</param>
    ''' <param name="value">The value of the channel</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function setDA(ByVal channel As Integer, ByVal value As Integer) As Boolean
        Dim response As String = ""
        Return sendCommand("SETDA " & channel & " " & value, response)
    End Function



    ''' <summary>
    ''' IOconfig() will set the I/O channel's configuration (input or output). You need to pass a string
    ''' to this function. This string must contain 16 bits (1 or 0). An 1 indicates a channel is an output,
    ''' a 0 indicates a channel is an input
    ''' </summary>
    ''' <param name="configStr">The configuration bit pattern</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function IOconfig(ByVal configStr As String) As Boolean
        If configStr.Length = 16 Then
            Dim response As String = ""
            If sendCommand("IOCONFIG " & configStr, response) Then
                Return True
            Else
                Return False
            End If
        Else
            Throw New Exception("configuration string has the wrong length!")
        End If
    End Function



    ''' <summary>
    ''' NOP() will send a 'no operation' command to the server. This can be used to keep the
    ''' server connections alive.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function NOP() As Boolean
        Dim response As String = ""
        Return sendCommand("NOP", response)
    End Function



    ''' <summary>
    ''' getADchannel() will read the value of A/D channel 1
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <param name="value"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function getADchannel(ByVal channel As Integer, ByRef value As Integer) As Boolean
        Dim response As String = ""

        If sendCommand("GETAD " & channel, response) Then
            ' Only proceed if the function succeeded
            Dim segments() As String = response.Split(" ")

            Dim intValue As Integer
            If Integer.TryParse(segments(1), intValue) Then
                ' The intvalue could be processed
                value = intValue
                Return True
            Else
                Return False
            End If
        Else
            Return False
        End If
    End Function




    ''' <summary>
    ''' getAllAD() will read the value of all A/D channels. The resulting data is available in the byref variables.
    ''' They should only be consulted if the getAllAD() function returned a 'true'.
    ''' </summary>
    ''' <param name="ad1">a/d channel 1</param>
    ''' <param name="ad2">a/d channel 2</param>
    ''' <param name="ad3">a/d channel 3</param>
    ''' <param name="ad4">a/d channel 4</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function getAllAD(ByRef ad1 As Integer, ByRef ad2 As Integer, ByRef ad3 As Integer, ByRef ad4 As Integer) As Boolean
        Dim response As String = ""

        If sendCommand("GETALLAD", response) Then
            ' Only proceed if the function succeeded
            Dim segments() As String = response.Split(" ")
            Dim intValue As Integer

            ' loop through the 4 returned codes
            For i As Integer = 1 To 4
                If Integer.TryParse(segments(i), intValue) Then
                    ' The intvalue could be processed
                    Select Case i
                        Case 1
                            ad1 = intValue
                        Case 2
                            ad2 = intValue
                        Case 3
                            ad3 = intValue
                        Case 4
                            ad4 = intValue
                    End Select
                Else
                    Return False
                End If
            Next

            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' setInitialized() will set the initialisation flag on the server.
    ''' </summary>
    ''' <param name="initialized">the desired state of the initialisation flag (true/false)</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function setInitialized(ByVal initialized As Boolean) As Boolean
        Dim initString As String
        Dim response As String = ""

        If initialized Then
            initString = "INIT 1"
        Else
            initString = "INIT 0"
        End If

        Return sendCommand(initString, response)
    End Function



    ''' <summary>
    ''' getInitialized() will get the initialisation flag from the server. This flag is passed through
    ''' the byref variable isInitialized.
    ''' </summary>
    ''' <param name="isInitialized">this variable indicates if the initialisation is done or not</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function getInitialized(ByRef isInitialized As Boolean) As Boolean
        Dim response As String = ""

        If sendCommand("INIT 2", response) Then
            Dim segments() As String = response.Split(" ")
            Dim intValue As Integer
            If (Integer.TryParse(segments(1), intValue)) Then
                If intValue = 1 Then
                    isInitialized = True
                Else
                    isInitialized = False
                End If

                Return True
            Else
                Return False
            End If
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' setDAC() will transmit a value for the DAC channel.
    ''' </summary>
    ''' <param name="channel">the channel to address</param>
    ''' <param name="value">the value the channel should get</param>
    ''' <returns>A boolean to indicate if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function setDAC(ByVal channel As Integer, ByVal value As Integer) As Boolean
        Dim response As String = ""
        Return sendCommand("SETDAC " & channel & " " & value, response)
    End Function



    ''' <summary>
    ''' setIOchannel() will set the specified I/O channel. The server response is ignored.
    ''' </summary>
    ''' <param name="channel">The channel to set</param>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function setIOchannel(ByVal channel As Integer) As Boolean
        Dim response As String = ""
        Return sendCommand("SETIO " & channel, response)
    End Function



    ''' <summary>
    ''' clearIOchannel() will set the specified I/O channel. The server response is ignored.
    ''' </summary>
    ''' <param name="channel">The channel to clear</param>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function clearIOchannel(ByVal channel As Integer) As Boolean
        Dim response As String = ""
        Return sendCommand("CLEARIO " & channel, response)
    End Function



    ''' <summary>
    ''' setIOchannel() will set the specified I/O channel. The server response is processed into the 
    ''' IOdata array.
    ''' </summary>
    ''' <param name="channel">The channel to clear</param>
    ''' <param name="IOdata">The IOdata array to store the resulting data. This array is passed ByRef. It is
    ''' not neccesary to dimension the array before calling this function.</param>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function setIOchannel(ByVal channel As Integer, ByRef IOdata() As Boolean) As Boolean
        Dim response As String = ""
        Dim IOstring As String = ""
        Dim serverResponse As Boolean

        serverResponse = sendCommand("SETIO " & channel, IOstring)

        If serverResponse Then
            getIOconfigFromServerResponse(IOstring, IOdata)
            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' clearIOchannel() will clear the specified I/O channel. The server response is processed into the 
    ''' IOdata array.
    ''' </summary>
    ''' <param name="channel">The channel to clear</param>
    ''' <param name="IOdata">The IOdata array to store the resulting data. This array is passed ByRef. It is
    ''' not neccesary to dimension the array before calling this function.</param>
    ''' <returns>A boolean indicating if the function succeeded</returns>
    ''' <remarks></remarks>
    Public Function clearIOchannel(ByVal channel As Integer, ByRef IOdata() As Boolean) As Boolean
        Dim response As String = ""
        Dim IOstring As String = ""
        Dim serverResponse As Boolean

        serverResponse = sendCommand("CLEARIO " & channel, IOstring)

        If serverResponse Then
            getIOconfigFromServerResponse(IOstring, IOdata)
            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' getAllIO() will retrieve the state of all I/O channels. The states are stored into the IOdata array,
    ''' which is passed by reference.
    ''' </summary>
    ''' <param name="IOdata">The IOdata array to store the resulting data. This array is passed ByRef. It is
    ''' not neccesary to dimension the array before calling this function.</param>
    ''' <returns>A boolean indicating if the function succeeded. This happens if 1) the data could be transmitted,
    ''' 2) a response was received and 3) the response matches the required bitformat.</returns>
    ''' <remarks></remarks>
    Public Function getAllIO(ByRef IOdata() As Boolean) As Boolean
        Dim IOstring As String = ""
        Dim serverResponse As Boolean

        serverResponse = sendCommand("GETALLIO", IOstring)

        If serverResponse Then
            Return getIOconfigFromServerResponse(IOstring, IOdata)
        Else
            Return False
        End If
    End Function

#End Region





#Region "Network Layer"
    ''' <summary>
    ''' getIOconfigFromServerResponse() will analyze the server response to see if it contains a valid row of 16 bits.
    ''' These bits will then be converted to the IOdata array.
    ''' </summary>
    ''' <param name="response">The response string to analyze</param>
    ''' <param name="IOdata">The IOdata array to store the resulting data. This array is passed ByRef. It is
    ''' not neccesary to dimension the array before calling this function.</param>
    ''' <returns>A boolean indicating if the server response was valid or not. A valid server response contains the
    ''' text +OK followed by 16 bits, representing the several I/O channels.</returns>
    ''' <remarks></remarks>
    Public Function getIOconfigFromServerResponse(ByRef response As String, ByRef IOdata() As Boolean) As Boolean
        If Regex.IsMatch(response, "^\+OK [0-1]{16}") Then
            ReDim IOdata(16)
            Dim bitstream As String = Mid(response, 5, 16)

            For i = 1 To 16
                Dim curvalue As String = Mid(bitstream, i, 1)
                If curvalue = "1" Then
                    IOdata(i - 1) = True
                Else
                    IOdata(i - 1) = False
                End If
            Next

            Return True
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' sendCommand will transport a new command over the network. The command should never take more than
    ''' one second to complete, otherwise an Exception will be generated.
    ''' </summary>
    ''' <param name="dataIn">the String to be transmitted</param>
    ''' <param name="dataOut">the response String from the server</param>
    ''' <returns>A boolean to indicate if the transmission succeeded or not.</returns>
    ''' <remarks></remarks>
    Private Function sendCommand(ByRef dataIn As String, ByRef dataOut As String) As Boolean

        Dim result As Boolean

        ' Do not proceed if the sendCommand is already busy (from another thread)
        If isSending Then Return False

        Try
            isSending = True
            networkStream.ReadTimeout = 1000
            networkStream.WriteTimeout = 1000
            Debug.Print("sendCommand: '" & dataIn & "'")
            Dim sendBytes As [Byte]() = Encoding.UTF8.GetBytes(dataIn & vbCrLf)
            networkStream.Write(sendBytes, 0, sendBytes.Length)
            ' Read the NetworkStream into a byte buffer.
            Dim bytes(tcpClient.ReceiveBufferSize) As Byte
            networkStream.Read(bytes, 0, CInt(tcpClient.ReceiveBufferSize))
            dataOut = Encoding.ASCII.GetString(bytes)
            isSending = False

            If Regex.IsMatch(dataOut, "^\+OK") Then
                result = True
            Else
                result = False
            End If

        Catch ex As Exception
            _lastException = ex
            isSending = False
            If isConnected = False Then RaiseEvent noConnectionError()
            result = False
        End Try

        isSending = False
        Return result
    End Function



    ''' <summary>
    ''' receiveCommand will receive the contents of the server. It should only be used to retrieve the first command
    ''' </summary>
    ''' <param name="dataOut">the response String from the server</param>
    ''' <returns>A boolean to indicate if the transmission succeeded or not.</returns>
    ''' <remarks></remarks>
    Private Function receiveCommand(ByRef dataOut As String) As Boolean
        Try

            ' Read the NetworkStream into a byte buffer.
            Dim bytes(tcpClient.ReceiveBufferSize) As Byte
            networkStream.Read(bytes, 0, CInt(tcpClient.ReceiveBufferSize))
            dataOut = Encoding.ASCII.GetString(bytes)
            Debug.Print("receiveCommand: " & dataOut)
            Return True
        Catch ex As Exception
            _lastException = ex
            Return False
        End Try
    End Function



    ''' <summary>
    ''' The connect method is used to establish a connection with the server.
    ''' </summary>
    ''' <param name="server">the host or ip address of the server</param>
    ''' <param name="port">the port of the server</param>
    ''' <param name="password">the password for the connection</param>
    ''' <returns>a boolean to indicate if the connection succeeded. The reason for a failed connection can be 
    ''' retrieved through the lastException property.</returns>
    ''' <remarks></remarks>
    Public Function connect(ByVal server As String, ByVal port As Integer, ByVal password As String) As Boolean
        Try
            ' Connect to the server and build the networkStream object
            tcpClient = New System.Net.Sockets.TcpClient()
            tcpClient.SendTimeout = 3000
            tcpClient.ReceiveTimeout = 3000
            tcpClient.Connect(server, port)
            networkStream = tcpClient.GetStream()

            ' We give the server 2 seconds initial response time
            networkStream.ReadTimeout = 2000
            networkStream.WriteTimeout = 2000


            If networkStream.CanWrite And networkStream.CanRead Then
                isConnected = True

                ' Read the inital data
                Dim initialData As String = ""
                receiveCommand(initialData)

                ' identify the server
                If Regex.IsMatch(initialData, "^\+OK Fez-to-K8000 TCP server") Then
                    ' Connection identified; Send the password
                    If sendCommand("PASS " & password, initialData) = False Then
                        disconnect()
                        Throw New Exception("The entered password was invalid!")
                    End If
                Else
                    disconnect()
                    Throw New Exception("Error: Unknown server version!")
                End If

                networkStream.ReadTimeout = 1000
                networkStream.WriteTimeout = 1000
            Else
                isConnected = False
            End If

            Return isConnected

        Catch ex As Exception
            _lastException = ex
            Return False
        End Try

    End Function



    ''' <summary>
    ''' The function disconnect will disconnect the client from the K8000 server. Any errors will
    ''' be ignored.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub disconnect()
        Try
            isConnected = False
            networkStream.Close()
            tcpClient.Close()
        Catch ex As Exception
            ' We don't need any error messages here
            ' Do *NOT* set the _lastException variable here, or other useful information might get lost 
        End Try
    End Sub
#End Region



    ''' <summary>
    ''' The constructor of the class defines a background timer that is checked every 30 seconds
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        controller = New Timer()
        controller.Enabled = True
        controller.Interval = 30000
    End Sub



    ''' <summary>
    ''' controller_Tick occurs every 30 seconds
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub controller_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles controller.Tick
        ' First we check if the client is still connected. If not, the clientDisconnected() event should be called
        ' Then we check the last transmission date: we send a NOP is the connection is about to timeout.
        If isConnected Then
            isConnected = tcpClient.Connected
            If isConnected = False Then
                ' Connection is closed since the last check
                RaiseEvent clientDisconnected()
            Else
                ' Send a NOP command
                If isSending = False Then NOP()
            End If
        End If
    End Sub
End Class
